#!/bin/env python

from datetime import datetime, timedelta
import sets

PAST=datetime.min
FUTURE=datetime.max

class Event:
    "Base class for all event classes."
    
    def __init__(self, start, end):
        self.start=start
        self.end=end

class EventSource:
    """Base class for all event source classes, subclasses should
    provide events from database, wiki or any other sources."""
    
    def getEvents(self, start, end):
        return None

def _cmpDates(d1, d2):
    return cmp(d1,d2)

def _cmpEvents(ev1, ev2, byEnd):
    if not byEnd:
        return _cmpDates(ev1.start, ev2.start) or \
               _cmpDates(ev1.end, ev2.end) or \
               cmp(ev1.__class__, ev2.__class__) or\
               cmp(ev1, ev2)
    else:
        return _cmpDates(ev1.end, ev2.end) or \
               _cmpDates(ev1.start, ev2.start) or \
               cmp(ev1.__class__, ev2.__class__) or\
               cmp(ev1, ev2)

def _search(lst, elem, cmp_, from_, to):
    "Binary search on sorted list."    
    if to==from_: return to
    cur = (to+from_)/2
    c = cmp_(lst[cur], elem)
    if c>=0:
        return _search(lst, elem, cmp_, from_, cur)
    else:
        return _search(lst, elem, cmp_, cur+1, to)

class Events:
    """Holds and manages events from several EventSource objects.

    Takes several EventSource objects and constructs two lists of all
    events, sorted by start and end."""
    
    def __init__(self, start, end, eventSources):
        evts = []
        for esource in eventSources:
            evts.extend(esource.getEvents(start, end))

        self.evtsAsc = list(evts)
        self.evtsAsc.sort(lambda x,y: _cmpEvents(x,y,False))
        self.evtsDsc = list(evts)
        self.evtsDsc.sort(lambda x,y: -_cmpEvents(x,y,True))

    def getEvents(self, from_, to):
        "Returns sets.Set of events in given period."
        cmpStart = lambda ev, d: cmp(ev.start, d)
        cmpEnd = lambda ev, d: -cmp(ev.end, d)
        
        iLstAsc = _search(self.evtsAsc, to, cmpStart, 0, len(self.evtsAsc))
        iLstDsc = _search(self.evtsDsc, from_, cmpEnd, 0, len(self.evtsDsc))

        evts = sets.Set(self.evtsAsc[:iLstAsc]) & sets.Set(self.evtsDsc[:iLstDsc])
        return evts

    def __str__(self):
        return "by start: " + str([str(ev) for ev in self.evtsAsc]) \
               + "\nby end: " + str([str(ev) for ev in self.evtsDsc])

class EventIterator:
    "Helper class to iterate over Events object events."
    
    def __init__(self, events, start, end, step):
        self.events=events
        self.start=start
        self.end=end
        self.step=step
        self.curStart=None
        self.curEnd=None
        self.curEvents=None

    def __iter__(self):
        return self

    def next(self):
        """Returns sets.Set of events for next period (date).

        Set returned must not be modified or it may break iteration."""
        if self.curEvents==None:
            self.curStart=self.start
            self.curEnd=self.start+self.step
            self.curEvents=self.events.getEvents(self.curStart, self.curEnd)
        else:
            # TODO it is possible to optimize this code
            #  if use self.events.evts[AD]sc lists directly
            self.curStart=self.curEnd
            if self.curStart>=self.end:
                raise StopIteration()
            self.curEnd=self.curStart+self.step
            self.curEvents=self.events.getEvents(self.curStart, self.curEnd)
            
        return self.curEvents

class EventTable:
    """Provides table data structure for events with rows, columns,
    row groups and special column group.

    This is primarly helper class for HTML rendering, but may be used
    in other places. Do not treat row or column as HTML row or column,
    table may be actually inversed, so <TR> may corespond to table columns
    and <TD> to table rows."""
    
    def __init__(self, events, start, end, step, hasColGroup=False):
        """events - Events object, start - datetime object, end -
        datetime object, step - timedelta object defining time change
        when going to the next column, hasColGroup - if there is
        special column group, which holds events spanning several
        rows.
        """
        self.events = events
        self.start = start
        self.end = end
        self.step = step
        self.hasColGroup = hasColGroup

        self._createRows()

    def isNewRow(self, end):
        """Must be defined by subclasses. Return True if new row to be
        started with next step."""
        return False

    def extractColIndex(self, d):
        """Must be defined by subclasses. Returns column index of date
        object."""
        raise Exception("Must be overriden")

    def isColGroupEvent(self, ev):
        """Must be defined by subclasses. Return True if this event
        belongs to special column group."""
        raise Exception("Must be overriden")

    def get(self, i, j): # irow, icol
        "Returns event in given cell."
        row = self.rows[i]
        return self.getCol(row, j)

    def getCol(self, row, j): # irow, icol
        "Returns event in given cell of row."
        for val in row:
            ev, from_, to = val
            if j>=from_ and j<to:
                return val
            
        return None

    def _createRows(self):
        "Actual table creation method."
        self.rows = []
        self.groups = []
        self.colGroup = []
        
        curRowIndex = 0

        iEvts = EventIterator(self.events, self.start, self.end, self.step)

        for evts in iEvts:
            # find new events
            newEvts = sets.Set(evts)
            for row in self.rows[curRowIndex:]:
                newEvts.discard(row[-1][0])
            for col in self.colGroup:
                newEvts.discard(col[-1][0])

            # add new events
            for newEvt in newEvts:
                if self.hasColGroup and self.isColGroupEvent(newEvt):
                    # add event to column group
                    for col in self.colGroup:
                        if not col[-1][2] == None:
                            col.append( (newEvt, len(self.groups), None) )
                            break
                    else:
                        self.colGroup.append( [(newEvt, len(self.groups), None)] )
                        
                else:
                    # add event to rows
                    for row in self.rows[curRowIndex:]:
                        if not row[-1][2] == None:
                            row.append( (newEvt, self.extractColIndex(iEvts.curStart), None) )
                            break
                    else:
                        self.rows.append( [(newEvt, self.extractColIndex(iEvts.curStart), None)] )

            # find ending events
            for row in self.rows[curRowIndex:]:
                evt, from_, to = row[-1]
                if to == None:
                    if evt.end <= iEvts.curEnd:
                        ind = self.extractColIndex(iEvts.curStart)+1
                        row[-1] = (evt, from_, ind)

            # end row if needed
            if self.isNewRow(iEvts.curEnd) or iEvts.curEnd==iEvts.end:
                # end all events
                for row in self.rows[curRowIndex:]:
                    evt, from_, to = row[-1]
                    if to == None:
                        # this event spans several groups
                        # just close it (it will be added again to the next group)
                        ind = self.extractColIndex(iEvts.curStart)+1
                        row[-1] = (evt, from_, ind)

                for col in self.colGroup:
                    evt, from_, to = col[-1]
                    if to == None:
                        if evt.end <= iEvts.curEnd or iEvts.curEnd==iEvts.end:
                            col[-1] = (evt, from_, len(self.groups)+1)

                # end this and start new row
                group = []
                group.extend(self.rows[curRowIndex:])
                self.groups.append(group)
                curRowIndex = len(self.rows)


        # count number of columns for table
        ncols = 0
        cur = self.start
        while cur<self.end:
            ncols = max(ncols, self.extractColIndex(cur)+1)
            cur = cur + self.step

        self.nrows = len(self.rows)
        self.ncols = ncols
