#!/bin/env python
"""Simple HTML parser and renderer module, allowing Python within HTML.

This module intended for simple HTML templating and rendering."""

from HTMLParser import HTMLParser
import sys
from StringIO import StringIO

# ---------- Parser ----------

class Parser(HTMLParser):
    """HTML parser that constructs tree of tags and raw text from HTML input.
    
Processing instructions are given within HTML template in form of <py
code='py:...'> tags and <anytag anyparam='py:...'> parameters. All
tags are compiled using Python compiler package and stored in tuples
with other text.

Use self.feed(string) to generate tree from HTML string which is
stored to self.html. Output can be passed further to Renderer class."""

    CODE_TAG_NAME="py"

    def __init__(self):
        HTMLParser.__init__(self)
        self.errFileName = "errors"
        self.html = [] # 'tree-like' result list of html strings and tuples(tag elements)
        self.__bodies = [self.html] # stack of current bodies during parsing

    def handle_starttag(self, tag, attrs):
        self.__openTag(tag, attrs)

    def handle_startendtag(self, tag, attrs):
        self.__appendTag((tag, attrs, None), False)

    def handle_endtag(self, tag):
        self.__closeTag(tag)

    def handle_data(self, data):
        self.__appendString(data)

    def handle_pi(self, data):
        pass

    def handle_entityref(self, name):
        self.__appendString("&%s;" % name)
            
    def __appendString(self, str_):
        """Append string to current body concatenating with last string
        if posible."""
        curBody = self.__bodies[-1]

        if len(curBody)==0:
            curBody.append(str_)
            return

        if type(curBody[-1])==str:
            curBody[-1] = "".join((curBody[-1], str_))
        else:
            curBody.append(str_)

    def __openTag(self, tagName, attrs):
        "Create new tuple and add to last body in output tree."
        tag = (tagName, attrs,[])
        self.__bodies[-1].append(tag)
        self.__bodies.append(tag[2])

    def __closeTag(self, tagName):
        """Remove all bodies from body stack until matching tag name
        is met.

        Close all tuples and concatenate strings where posible."""
        
        # remove bodies, check for tag name match and concat if possible
        while len(self.__bodies) > 1: # there always must be 'file' body as first element
            del self.__bodies[-1]

            tag = self.__bodies[-1][-1]

            del self.__bodies[-1][-1]  # parent body last element is tag to remove
            self.__appendTag(tag,
                             hasEndTag=(tag[0]==tagName))

            if tag[0]==tagName: # tag name matches
                return

        # hm, no tag match throw exception
        raise Exception("No start tag found for %s" % tagName)

    def __appendTag(self, tag, hasEndTag):
        """Working method which does all the generating work.

        Compiles tags and parameters, concatenates strings, creates final tuples."""
        tagName, attrs, body = tag

        # divide into processed attributes and literal attributes
        ltrAttrs = []
        prcsAttrs = []

        import compiler
        for a in (attrs or []):
            if a[1] and a[1].startswith("py:"):
                prcsAttrs.append(
                    (a[0], compiler.compile(a[1][3:],
                                            self.errFileName,
                                            tagName==self.CODE_TAG_NAME and a[0]=="code" and "single" or "eval")))
            else:
                ltrAttrs.append(a)

        # if no processing attrs then add as string (optimization)
        if len(prcsAttrs)==0:
            # add start tag as string
            argLst = map(lambda x: '%s="%s"' % x, ltrAttrs)
            lst = [tagName]
            lst.extend(argLst)
            str_ = "<%s%s>" % (" ".join(lst), body==None and "/" or "")
            self.__appendString(str_)
            # add body first element as string if possible
            if body!=None and len(body)>0 and type(body[0])==str:
                self.__appendString(body[0])
                self.__bodies[-1].extend(body[1:])
            elif body!=None:
                self.__bodies[-1].extend(body)
            # add end tag
            if hasEndTag:
                self.__appendString("</%s>"%tagName)
                
        else: # add as tuple
            args = []
            args.extend(prcsAttrs)
            args.extend(ltrAttrs)
            self.__bodies[-1].append((tagName, args, body))


# ---------- Renderer ----------

class Renderer:
    """Base class for all renderers that generate HTML from Parser
    output.

    Renderer walks tuple tree and writes it to formatter, executing
    compiled Python code."""

    def __init__(self, template, parent=None):
        """Constructor.

        template - Parser output tree, parent - parent Renderer."""
        self.template = template
        self.formatter = None
        if parent == None:
            parent = self
        self.parent = parent

    def render(self, formatter, template=None):
        """Entry method for rendering component, it should be invoked
        from outside of component.

        formatter - where output should be written, template -
        alternate tree used instead of original (this is convinience
        parameter and should be removed in the future)

        It should NOT be overriden, because it is intended to
        be unique for all components."""
        
        if template!=None:
            self.template = template

        self._renderComponent(formatter)

    def _renderComponent(self, formatter):
        """Actual component rendering method.

        Should not be called from any place except this component
        render method. Should be overriden, default implementation
        just calls renderBody() on parent component."""
        self.parent.renderBody(self.template, formatter)

    def renderBody(self, _body, formatter):
        """Actual working method which processes piece of given tuple tree.

        _body - tuple tree to be processed, formatter - where output is written.

        When used from _renderComponent() with self.template as _body
        should be called on self.parent component, becase body is
        contained within parent and user expects it to be in parent
        context."""
        renderBody = self.renderBody
        
        for comp in _body:
            if type(comp)==str:
                formatter.write(comp)
                
            elif type(comp)==tuple:
                tagName, attrs, body = comp
                
                if tagName!="py":
                    attrLst = filter(lambda x: x[1]!=None,
                                     [(a, type(v)==str and v or eval(v))
                                      for (a,v) in attrs])
                    attrLst = ['%s="%s"' % (a,v) for a,v in attrLst]
                    lst = [tagName] + attrLst;
                    
                    formatter.write("<%s>" % " ".join(lst))
                    renderBody(body, formatter)
                    formatter.write("</%s>" % tagName)
                    
                else:
                    eval(attrs[0][1])

class DelegateFormatter:
    "Delegates output to internal formatter."
    
    def __init__(self):
        self.formatter = None
    
    def _null(self, *args, **kwds):
        pass
    
    def __getattr__(self, name):
        if self.formatter == None:
            return self._null
        else:
            return getattr(self.formatter, name)

class Buffers (Renderer):
    """Special renderer which collects output to buffers and then
    flushes it by command.

    It contains list of anonymous and key buffers, last are created
    with renderBlock() method."""
    
    def __init__(self, parent):
        Renderer.__init__(self, None, parent)
        self.delegateFormatter = DelegateFormatter()
        self.iostrings = []
        self.blocks = dict()
        self.onceRendered = False

    def _renderComponent(self, formatter):
        "Renders only once and all subsequent output is discarded until flush() is called."
        
        self.outFormatter = formatter
        
        if not self.onceRendered:
            self.iostrings.append(StringIO())
            self.delegateFormatter.formatter = self.iostrings[-1]
        else:
            self.delegateFormatter.formatter = None

        self.parent.renderBody(self.template, self.delegateFormatter)

        self.onceRendered = True

    def renderBlock(self, blockKey, body):
        """Special method that renders body to special buffer defined by blockKey.

        This method should be used from within HTML template. Called
        with the same blockKey it will append output to existing
        buffer. If buffer with this key does not exist, it will be
        created and appended to the end of buffers' list."""
        block = self.blocks.get(blockKey)
        if block==None:
            block=StringIO()
            self.blocks[blockKey]=block
            self.iostrings.append(block)
            self.delegateFormatter.formatter = self.iostrings[-1]

        self.parent.renderBody(body, block)

        if not self.onceRendered:
            block=StringIO()
            self.iostrings.append(block)
            self.delegateFormatter.formatter = self.iostrings[-1]
        else:
            self.delegateFormatter.formatter = None            

    def flush(self):
        """Flushes all buffers to formatter given in render() method.

        It also forgets about its render() calls, ie new call to this
        method do not discard output, but stores to first anonymous
        buffer."""
        
        for iostr in self.iostrings:
            self.outFormatter.write(iostr.getvalue())

        self.iostrings = []
        self.blocks = dict()
        self.onceRendered = False

# ---------- Main ----------

if __name__=="__main__":
    p = Parser()
    p.feed(open("test.html.tmpl").read())

    print p.html

    class TestRenderer(Renderer):
        def __init__(self, template):
            Renderer.__init__(self, template)
            self.buffers = Buffers(self)
        
    r = TestRenderer(template=p.html)
    r.render(formatter=sys.stdout)

